MODULE realdatalayout

IMPLICIT NONE

    INTEGER, PARAMETER :: alongX=1, alongY=2

    TYPE REAL_DATA_LAYOUT
        INTEGER :: nx         !< Global grid size for the datas
        INTEGER :: ny         !< Global grid size for the datas
        ! curlayout should be alongX or alongY
        INTEGER :: curlayout  !< should be alongX or alongY or alongZ
        INTEGER :: xmin       !< start X position in the global array for the current layout
        INTEGER :: xmax       !< final X position in the global array for the current layout
        INTEGER :: ymin       !< start Y position in the global array for the current layout
        INTEGER :: ymax       !< final Y position in the global array for the current layout

        !> The datas. The array size is (xmin:xmax,ymin:ymax)
        INTEGER, DIMENSION(:,:), POINTER :: values => NULL()

        !> A ncpus x 2 array: ymin and ymax values for each cpu when the layout is along X axis
        INTEGER, DIMENSION(:,:),POINTER :: Y4layoutOnX => NULL()
        !> A ncpus x 2 array: xmin and xmax values for each cpu when the layout is along Y axis
        INTEGER, DIMENSION(:,:),POINTER :: X4layoutOnY => NULL()
    END TYPE REAL_DATA_LAYOUT

CONTAINS
    LOGICAL FUNCTION real_initDataLayout(val,nx,ny,nbcpus,me,how)
       IMPLICIT NONE
       INTEGER, INTENT(IN)           :: nx,ny
       TYPE(REAL_DATA_LAYOUT),INTENT(INOUT) :: val
       INTEGER, INTENT(IN)           :: nbcpus,me
       INTEGER, INTENT(IN),OPTIONAL  :: how   !default to alongX

       INTEGER :: ires

       real_initDataLayout=.FALSE.

       !if something is allocated, deallocate it
       IF(ASSOCIATED(val%values)) DEALLOCATE(val%values)
       IF(ASSOCIATED(val%Y4layoutOnX)) DEALLOCATE(val%Y4layoutOnX)
       IF(ASSOCIATED(val%X4layoutOnY)) DEALLOCATE(val%X4layoutOnY)

       ! global dimensions
       val%nx=nx
       val%ny=ny


      !get the data organisation on the cpus
      IF (.NOT. ( &
          & initLayoutArray(val%Y4layoutOnX,nbcpus,ny)&
          & .AND. &
          & initLayoutArray(val%X4layoutOnY,nbcpus,nx)&
          )) THEN
          WRITE(6,'(a,i0,a)')"ERROR on process ",me,": not enought memory!"
          RETURN
      ENDIF

        ! get the initial datalayout
      IF (present(how)) THEN
        IF (how.EQ.alongX .OR. how .EQ. alongY) THEN
            val%curlayout=how
        ELSE
            WRITE(6,'(a,i0,a,i0)')"ERROR on process ",me,": unknown layout ",how
            WRITE(6,'(a)')        "      Should be in [1,2]"
        ENDIF
      ELSE
        val%curlayout=alongX
      ENDIF

      !Allocate data array, we know curlayout is right so no default case.
      SELECT CASE (val%curlayout)
      CASE (alongX)
        val%xmin=1
        val%xmax=nx
        val%ymin=val%Y4layoutOnX(me+1,1)
        val%ymax=val%Y4layoutOnX(me+1,2)
      CASE (alongY)
        val%xmin=val%X4layoutOnY(me+1,1)
        val%xmax=val%X4layoutOnY(me+1,2)
        val%ymin=1
        val%ymax=ny
      END SELECT
      ALLOCATE(val%values(val%xmin:val%xmax ,val%ymin:val%ymax),stat=ires)
      IF (ires .NE.0) THEN
        WRITE(6,'(a,i0,a)')"ERROR on process ",me,": not enought memory for datas!"
        RETURN
      ELSE
        val%values=0
        real_initDataLayout=.TRUE.
      ENDIF
     
      RETURN

    END FUNCTION real_initDataLayout

    !===========================================
    LOGICAL FUNCTION initLayoutArray(layout,ncpus,nbpt)
    !===========================================

      IMPLICIT NONE
      INTEGER, INTENT(IN)::ncpus,nbpt
      INTEGER, DIMENSION(:,:),POINTER, INTENT(INOUT) ::layout

      INTEGER:: res,i,start,offset

      initLayoutArray=.FALSE.    
      if (nbpt .LT. ncpus) then
         WRITE(6,'(a)')"ERROR in initLayoutArray: less points than cpus!"
         RETURN
      ENDIF

      ALLOCATE(layout(ncpus,2),stat=res)
      IF (res.NE.0) THEN
         WRITE(6,'(a)')"ERROR in initLayoutArray: not enought memory!"
         RETURN
      ENDIF

      res=mod(nbpt,ncpus)
      offset=nbpt/ncpus
      start=1
      DO i=1,ncpus
         layout(i,1)=start
         IF (res.GT.0) THEN
            layout(i,2)=start+offset
            res=res-1
         ELSE
            layout(i,2)=start+offset-1
         ENDIF
         start=layout(i,2)+1
      ENDDO
      initLayoutArray=.TRUE.
    END FUNCTION initLayoutArray


    !===========================================
    SUBROUTINE real_deleteDataLayout(val)
    !===========================================
      TYPE(REAL_DATA_LAYOUT),INTENT(INOUT) :: val

      IF(ASSOCIATED(val%values))       DEALLOCATE(val%values)
      IF(ASSOCIATED(val%Y4layoutOnX))  DEALLOCATE(val%Y4layoutOnX)
      IF(ASSOCIATED(val%X4layoutOnY))  DEALLOCATE(val%X4layoutOnY)  
      val%values => NULL()
      val%Y4layoutOnX => NULL()
      val%X4layoutOnY => NULL()
      val%xmin=0
      val%xmax=-1
      val%ymin=0
      val%ymax=-1

    END SUBROUTINE real_deleteDataLayout

END MODULE realdatalayout
