MODULE realparallel
    USE realdatalayout
    USE mpi


    IMPLICIT NONE


    !private subroutines & functions
    PRIVATE real_AlongXtoAlongY
    PRIVATE real_AlongYtoAlongX

CONTAINS
  
!------------------------------------------------------------------------------
!> @author 
!> Patrick BEGOU
!
!> @details
!> This subroutine changes the shape of the REAL_DATA_LAYOUT variable: val. Datas are set
!> contiguous along the X axis and distributed on Y among the processes.
!> @param[in,out] val the REAL_DATA_LAYOUT variable to redistribute.
!> @param[in] spec_rank my rank in spec_communitor
!> @param[in] keepTypes, optional: keep the type created for the sub arrays in the datalayout for next use. Default is FALSE
!------------------------------------------------------------------------------
LOGICAL FUNCTION real_setAlongX(spec_rank,val)

    IMPLICIT NONE
    TYPE(REAL_DATA_LAYOUT),INTENT(INOUT) :: val
    INTEGER, INTENT(IN)             :: spec_rank


    real_setAlongX=.FALSE.

    SELECT CASE (val%curlayout)
    CASE (alongX)  
        real_setAlongX=.TRUE.
        RETURN
    CASE (alongY)
        real_setAlongX=real_AlongYtoAlongX(spec_rank,val) 
        RETURN
    END SELECT
    RETURN

END FUNCTION real_setAlongX

!------------------------------------------------------------------------------
!> @author 
!> Patrick BEGOU
!
!> @details
!> This subroutine changes the shape of the REAL_DATA_LAYOUT variable: val. Datas are set
!> contiguous along the Y axis and distributed on X among the processes.
!> @param[in,out] val the REAL_DATA_LAYOUT variable to redistribute.
!> @param[in] spec_rank my rank in spec_communitor
!> @param[in] keepTypes, optional: keep the type created for the sub arrays in the datalayout for next use. Default is FALSE
!------------------------------------------------------------------------------
LOGICAL FUNCTION real_setAlongY(spec_rank,val)

    IMPLICIT NONE
    TYPE(REAL_DATA_LAYOUT),INTENT(INOUT) :: val
    INTEGER, INTENT(IN)             :: spec_rank


    real_setAlongY=.FALSE.

    SELECT CASE (val%curlayout)
    CASE (alongX)  
        real_setAlongY=real_AlongXtoAlongY(spec_rank,val) 
        RETURN
    CASE (alongY)
        real_setAlongY=.TRUE.
        RETURN
    END SELECT
    RETURN

END FUNCTION real_setAlongY

!------------------------------------------------------------------------------
!> @author 
!> Patrick BEGOU
!
!> @details
!> This subroutine changes the shape of the REAL_DATA_LAYOUT variable: val. Datas are 
!> supposed to be contiguous along X and are reorganized to be contiguous along Y 
!> on the current processor. This version uses specific datatype.
!> @param[in,out] val the REAL_DATA_LAYOUT variable to redistribute.
!> @param[in] spec_rank my rank in spec_communitor
!> @param[in] keep_type_arrays if TRUE keep the arrays of subarray types in the datalayout for next use.
!------------------------------------------------------------------------------
  LOGICAL FUNCTION real_AlongXtoAlongY(spec_rank,val)
 
  IMPLICIT NONE
  TYPE(REAL_DATA_LAYOUT),INTENT(INOUT) :: val
  INTEGER, INTENT(IN)                  :: spec_rank
  

  INTEGER                            :: res,i,nbcpus,ierr
  INTEGER                            :: subArraySize(2), subArrayStart(2),fullArraySize(2)
  INTEGER, DIMENSION(:,:),POINTER    :: values => NULL()  
  INTEGER, DIMENSION(:), ALLOCATABLE :: sendcount, senddist, Bloc_onX4Y, Bloc_onY4X

  
  real_AlongXtoAlongY=.FALSE.

  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,nbcpus,ierr)


  ALLOCATE(Bloc_onX4Y(nbcpus),Bloc_onY4X(nbcpus),stat=res)
  IF(res.NE.0) THEN
     WRITE(6,'(a,i0,a)')"ERROR real_AlongXtoAlongY on process ",spec_rank,  &
                     &  ": not enought memory for internal arrays!"
     RETURN
  ENDIF

  !Creation des types MPI pour le setup along X dans le plan [XY]
  fullArraySize(1)=val%nx
  fullArraySize(2)=val%Y4layoutOnX(spec_rank+1,2)-val%Y4layoutOnX(spec_rank+1,1) + 1
  DO i=1,nbcpus
        subArraySize(1)=val%X4layoutonY(i,2) - val%X4layoutonY(i,1) + 1 
        subArraySize(2)=val%Y4layoutOnX(spec_rank+1,2)-val%Y4layoutOnX(spec_rank+1,1) + 1

        subArrayStart(1)=val%x4layoutony(i,1)-1
        subArrayStart(2)=0

        CALL MPI_type_Create_SubArray( 2,            &
                             &    fullArraySize,     &
                             &    subArraySize,      &
                             &    subArrayStart,     &
                             &    MPI_ORDER_FORTRAN, &
                             &    MPI_INTEGER,       &
                             &    Bloc_onX4Y(i), &
                             &    ierr               )
        CALL MPI_type_commit(Bloc_onX4Y(i),ierr)
     ENDDO

     !Creation des types MPI pour le setup along Y dans le plan [XY]
     fullArraySize(1)=val%X4layoutOnY(spec_rank+1,2)-val%X4layoutOnY(spec_rank+1,1) + 1
     fullArraySize(2)=val%ny
  
     DO i=1,nbcpus
        subArraySize(1)=val%X4layoutOnY(spec_rank+1,2)-val%X4layoutOnY(spec_rank+1,1) + 1
        subArraySize(2)=val%Y4layoutonX(i,2) - val%Y4layoutonX(i,1) + 1 

        subArrayStart(1)=0
        subArrayStart(2)=val%Y4layoutonX(i,1)-1

        CALL MPI_type_Create_SubArray( 2,            &
                             &    fullArraySize,     &
                             &    subArraySize,      &
                             &    subArrayStart,     &
                             &    MPI_ORDER_FORTRAN, &
                             &    MPI_INTEGER,       &
                             &    Bloc_onY4X(i), &
                             &    ierr               )
        CALL MPI_type_commit(Bloc_onY4X(i),ierr)
     ENDDO

!  Allocate receive buffer  along Y
  ALLOCATE(values(val%X4layoutOnY(spec_rank+1,1):val%X4layoutOnY(spec_rank+1,2), &
                & 1:val%ny), stat=res)
  IF(res.NE.0) THEN
       WRITE(6,'(a,i0,a)')"ERROR real_AlongXtoAlongY on process ",spec_rank, &
                        & ": not enought memory for datas!"
       WRITE(6,101)  spec_rank, & 
                  & val%X4layoutOnY(spec_rank+1,1),val%X4layoutOnY(spec_rank+1,2), &
                  & 1,val%ny, &
                  & " has FAILED"
101 format ("On process ",i0," allocation of values: [",i0,":",i0,",",i0,":",i0,"]",a)
       RETURN
  ENDIF

  ! now send & receive the datas with an ALL 2 ALL in the local plane
  ALLOCATE(sendcount(nbcpus), senddist(nbcpus),stat=res)
  IF(res.NE.0) THEN
     WRITE(6,'(a,i0,a)')"ERROR real_AlongXtoAlongY on process ",spec_rank, &
                      & ": not enought memory for internal arrays!"
     RETURN
  ENDIF
  sendcount(:)=1
  senddist(:)=0
  CALL MPI_ALLTOALLW(val%values,   &
             &     sendcount,      &
             &     senddist,       &
             &     Bloc_onX4Y,     &
             &     values,         &
             &     sendcount,      &
             &     senddist,       &
             &     Bloc_onY4X,     &
             &     MPI_COMM_WORLD, &
             &     ierr            )

  DEALLOCATE(sendcount,senddist)

  DO i=1,nbcpus
     CALL MPI_type_free(Bloc_onX4Y(i),ierr)
     CALL MPI_type_free(Bloc_onY4X(i),ierr)
  ENDDO
  DEALLOCATE(Bloc_onX4Y,Bloc_onY4X)

  !Update val structure pointing to this new array
  DEALLOCATE(val%values)
  val%values=>values
  values=>null()
  val%xmin=val%X4layoutOnY(spec_rank+1,1)
  val%xmax=val%X4layoutOnY(spec_rank+1,2)
  val%ymin=1
  val%ymax=val%ny
  val%curlayout=alongY
  
  real_AlongXtoAlongY=.TRUE.
  RETURN
  END FUNCTION real_AlongXtoAlongY
  
!> @author 
!> Patrick BEGOU
!
!> @details
!> This subroutine changes the shape of the REAL_DATA_LAYOUT variable: val. Datas are 
!> supposed to be contiguous along Y and are reorganized to be contiguous along X 
!> on the current processor. This version uses specific datatype.
!> @param[in,out] val the REAL_DATA_LAYOUT variable to redistribute.
!> @param[in] spec_rank my rank in spec_communitor
!> @param[in] keep_type_arrays if TRUE keep the arrays of subarray types in the datalayout for next use.
! See real_AlongXtoAlongY() for more details and comments.
!------------------------------------------------------------------------------
  LOGICAL FUNCTION real_AlongYtoAlongX(spec_rank,val)

  IMPLICIT NONE

  TYPE(REAL_DATA_LAYOUT),INTENT(INOUT) :: val
  INTEGER, INTENT(IN)                  :: spec_rank  

  INTEGER                            :: subArraySize(2), subArrayStart(2),fullArraySize(2)
  INTEGER, DIMENSION(:), ALLOCATABLE :: sendcount, senddist, Bloc_onX4Y, Bloc_onY4X
  INTEGER                            :: res,i,nbcpus,ierr
  INTEGER, DIMENSION(:,:),POINTER       :: values => NULL()     
  
  real_AlongYtoAlongX=.FALSE.
  CALL MPI_COMM_SIZE(MPI_COMM_WORLD,nbcpus,ierr)

  ALLOCATE(Bloc_onX4Y(nbcpus), Bloc_onY4X(nbcpus), stat=res)
  IF(res.NE.0) THEN
     WRITE(6,'(a,2(i0,a))')"ERROR real_AlongYtoAlongX on process ",spec_rank, &
                         & ": not enought memory (",nbcpus*4,") for internal arrays!"
     RETURN
  ENDIF

  !Creation des types MPI pour le setup along X dans le plan [XY]
  fullArraySize(1)=val%nx
  fullArraySize(2)=val%Y4layoutOnX(spec_rank+1,2)-val%Y4layoutOnX(spec_rank+1,1) + 1
  DO i=1,nbcpus
     subArraySize(1)=val%X4layoutonY(i,2) - val%X4layoutonY(i,1) + 1 
     subArraySize(2)=val%Y4layoutOnX(spec_rank+1,2)-val%Y4layoutOnX(spec_rank+1,1) + 1

     subArrayStart(1)=val%X4layoutonY(i,1)-1
     subArrayStart(2)=0

     CALL MPI_type_Create_SubArray( 2,            &
                          &    fullArraySize,     &
                          &    subArraySize,      &
                          &    subArrayStart,     &
                          &    MPI_ORDER_FORTRAN, &
                          &    MPI_INTEGER,       &
                          &    Bloc_onX4Y(i), &
                          &    ierr               )
     CALL MPI_type_commit(Bloc_onX4Y(i),ierr)
  ENDDO

     !Creation des types MPI pour le setup along Y dans le plan [XY]
  fullArraySize(1)=val%X4layoutOnY(spec_rank+1,2)-val%X4layoutOnY(spec_rank+1,1) + 1
  fullArraySize(2)=val%ny
  
  DO i=1,nbcpus
     subArraySize(1)=val%X4layoutOnY(spec_rank+1,2)-val%X4layoutOnY(spec_rank+1,1) + 1
     subArraySize(2)=val%Y4layoutonX(i,2) - val%Y4layoutonX(i,1) + 1 

     subArrayStart(1)=0
     subArrayStart(2)=val%Y4layoutonX(i,1)-1

     CALL MPI_type_Create_SubArray( 2,            &
                          &    fullArraySize,     &
                          &    subArraySize,      &
                          &    subArrayStart,     &
                          &    MPI_ORDER_FORTRAN, &
                          &    MPI_INTEGER,       &
                          &    Bloc_onY4X(i), &
                          &    ierr               )
     CALL MPI_type_commit(Bloc_onY4X(i),ierr)
  ENDDO


  !allocate receive buffer  along X
  ALLOCATE(values(1:val%nx, &
              & val%Y4layoutOnX(spec_rank+1,1):val%Y4layoutOnX(spec_rank+1,2)), &
              & stat=res)
  IF(res.NE.0) THEN
       WRITE(6,'(a,i0,a)')"ERROR real_AlongYtoAlongX on process ",spec_rank, &
                        & ": not enought memory for datas!"
       WRITE(6,101)  spec_rank, & 
                  & val%X4layoutOnY(spec_rank+1,1),val%X4layoutOnY(spec_rank+1,2), &
                  & 1,val%ny, &
                  & " has FAILED"
101 format ("On process ",i0," allocation of values: [",i0,":",i0,",",i0,":",i0,"]",a)
       RETURN
  ENDIF

  ! now send & receive the datas
  ALLOCATE(sendcount(nbcpus), senddist(nbcpus),stat=res)
  IF(res.NE.0) THEN
     WRITE(6,'(a,i0,a)')"ERROR real_AlongYtoAlongX on process ",spec_rank, &
                      & ": not enought memory ("                           &
        & ,nbcpus*4,") for internal sendcount/senddist arrays!"
     RETURN
  ENDIF
  sendcount(:)=1
  senddist(:)=0
  CALL MPI_ALLTOALLW(val%values,   &
             &     sendcount,      &
             &     senddist,       &
             &     Bloc_onY4X,     &
             &     values,         &
             &     sendcount,      &
             &     senddist,       &
             &     Bloc_onX4Y,     &
             &     MPI_COMM_WORLD, &
             &     ierr            )

  DEALLOCATE(sendcount,senddist)
  DO i=1,nbcpus
     CALL MPI_type_free(Bloc_onX4Y(i),ierr)
     CALL MPI_type_free(Bloc_onY4X(i),ierr)
  ENDDO
  DEALLOCATE(Bloc_onX4Y,Bloc_onY4X)

  !Update val structure pointing to this new array
  DEALLOCATE(val%values)
  val%values=>values
  values=>null()
  val%xmin=1
  val%xmax=val%nx
  val%ymin=val%Y4layoutOnX(spec_rank+1,1)
  val%ymax=val%Y4layoutOnX(spec_rank+1,2)
  val%curlayout=alongX
  
  real_AlongYtoAlongX=.TRUE.
  RETURN
  
  END FUNCTION real_AlongYtoAlongX


END MODULE realparallel
